from scipy.spatial import voronoi_plot_2d

from voronout.VoronoiDiagram import VoronoiDiagram
from voronout.Point import Point 
from voronout.VoronoiDiagramToJSON import toJson
from matplotlib import pyplot

import random

# basePoints are percentages that we scale up later
# looks good now - just fix a few bugs - division by 0, lines crossing over, edge case with the above three points - clean up tests, and we are good

basePoints = tuple((Point(x = random.random(), y = random.random()) for _ in range(15)))
print(tuple((repr(basePoint) for basePoint in basePoints)))

baseWidth = 600
baseHeight = 600

voronoiDiagram = VoronoiDiagram(basePoints = basePoints, planeWidth = baseWidth, planeHeight = baseHeight)

pyplot.ylim(bottom = baseWidth, top = 0)

for voronoiRegion in voronoiDiagram.voronoiRegions.values():
    for voronoiRegionEdge in voronoiRegion.edges:
        vertexIdentifier0 = voronoiRegionEdge.vertex0Id
        vertexIdentifier1 = voronoiRegionEdge.vertex1Id

        vertex0 = voronoiDiagram.vertices[vertexIdentifier0]
        vertex1 = voronoiDiagram.vertices[vertexIdentifier1]
        
        pyplot.plot([vertex0.x, vertex1.x], [vertex0.y, vertex1.y])
    
voronoi_plot_2d(voronoiDiagram._voronoiDiagram, show_points = True, show_vertices = True)
pyplot.show()

toJson(voronoiDiagram = voronoiDiagram, voronoiJsonPath = "voronoi.json")


from dataclasses import dataclass
from math import sqrt

import numpy as np

from .utils import boundValue

@dataclass(frozen=True)
class Point:
    x: float
    y: float

    def __init__(self, x, y):
        object.__setattr__(self, "x", boundValue(value = x))
        object.__setattr__(self, "y", boundValue(value = y))

    @staticmethod
    def distance(p1: Point, p2: Point) -> float:
        dx = p2.x - p1.x
        dy = p2.y - p1.y
        return boundValue(value = sqrt(pow(dx, 2) + pow(dy, 2)))

    @staticmethod
    def midpoint(p1: Point, p2: Point) -> Point:
        midpointX = boundValue(value = (p1.x + p2.x)/2)
        midpointY = boundValue(value = (p1.y + p2.y)/2)

        return Point(x = midpointX, y = midpointY)
    
    # For converting between 0, 0 (top-left) and 0, 0 (bottom-left).
    def convertPointBase(self) -> Point:
        return Point(x = self.x, y = boundValue(value = 1 - self.y))
    
    # For scaling a Point up when outputting it.
    def scale(self, widthScalar: float, heightScalar: float) -> Point:
        return Point(x = boundValue(value = self.x * widthScalar), y = boundValue(value = self.y * heightScalar))
    
    # For ease of JSON conversion.
    def __repr__(self) -> str:
        # -0.0 and 0.0 are equivalent in Python - no reason for JSON output of both not to uniformly be 0.0.
        x = abs(self.x) if self.x == 0.0 else self.x
        y = abs(self.y) if self.y == 0.0 else self.y

        return f'{{"x": {x}, "y": {y}}}'
    
    # For using Point in scipy.spatial methods that take ArrayLike parameters.
    def __array__(self, dtype=None, copy=None):
        return np.array(tuple((self.x, self.y)))
    
    # For checking that Point is in a given list.
    def __eq__(self, other):
        return self.x == other.x and self.y == other.y


from ..Point import Point

import json
import numpy as np

testPoint = Point(x = 0.3, y = 0.4)
testOtherPoint = Point(x = -0.1, y = 0.2)

def test_distance():
    distance = Point.distance(p1 = testPoint, p2 = testOtherPoint)
    # distance = sqrt((x2 - x1)^2 + (y2 - y1)^2)
    assert distance == 0.4472

def test_midpoint():
    midpoint = Point.midpoint(p1 = testPoint, p2 = testOtherPoint)

    # midpoint = ((x1 + x2)/2, (y1 + y2)/2)
    assert midpoint.x == 0.1
    assert midpoint.y == 0.3

def test_to_json():
    testPointJson = json.loads(repr(testPoint))

    assert len(testPointJson.keys()) == 2

    assert testPointJson["x"] == testPoint.x
    assert testPointJson["y"] == testPoint.y

def test_convert_point_base():
    pointBaseConverted = testPoint.convertPointBase()

    assert pointBaseConverted.x == testPoint.x
    assert pointBaseConverted.y == 1 - testPoint.y

def test_scale():
    widthScalar = 10
    heightScalar = 10

    pointScaled = testPoint.scale(widthScalar = widthScalar, heightScalar = heightScalar)
    assert pointScaled.x == testPoint.x * widthScalar
    assert pointScaled.y == testPoint.y * heightScalar

def test_to_array():
    npArray = np.array(testPoint)
    expectedNpArray = np.array(tuple((testPoint.x, testPoint.y)))
                               
    np.testing.assert_array_equal(npArray, expectedNpArray)


from ..VoronoiRegionData import VoronoiRegionData

from uuid import uuid4

from ...edges.VoronoiEdgeData import VoronoiEdgeData
from ...edges.VoronoiEdge import VoronoiEdge

from ...Point import Point

# voronoi_diagram_test implicitly covers the success case.
def test_not_adding_edge_already_present():
    testRegionData = VoronoiRegionData(siteId = uuid4())

    testPoint0 = Point(x = 0.1, y = 0.1)
    testPoint1 = Point(x = 0.1, y = 0.2)

    testEdge = VoronoiEdge(vertex0Id = uuid4(), vertex1Id = uuid4(), neighborSiteId = uuid4())
    testEdgeData = VoronoiEdgeData(vertex0 = testPoint0, vertex1 = testPoint1, edgeInIds = testEdge)

    assert testRegionData.addEdgesData(edgeVertex0Index = 0, edgeVertex1Index = 1, neighborSiteIndex = 1, edgeData = testEdgeData)

    # Make sure that it covers both possible ways the edge could be described - A->B and B->A.
    assert not testRegionData.addEdgesData(edgeVertex0Index = 0, edgeVertex1Index = 1, neighborSiteIndex = 1, edgeData = testEdgeData)
    assert not testRegionData.addEdgesData(edgeVertex0Index = 1, edgeVertex1Index = 0, neighborSiteIndex = 1, edgeData = testEdgeData)


from uuid import uuid4

from ..edges.VoronoiEdge import VoronoiEdge
from ..regions.VoronoiRegion import VoronoiRegion

import json

def _makeVoronoiEdge() -> VoronoiEdge:
    return VoronoiEdge(vertex0Id = uuid4(), vertex1Id = uuid4(), neighborSiteId = uuid4())

testVoronoiEdge = _makeVoronoiEdge()
testOtherVoronoiEdge = _makeVoronoiEdge()
testVoronoiRegion = VoronoiRegion(siteId = uuid4(), edges = tuple((testVoronoiEdge, testOtherVoronoiEdge)))

def test_neighbors():
    regionNeighbors = testVoronoiRegion.neighbors()
    assert tuple((testVoronoiEdge.neighborSiteId, testOtherVoronoiEdge.neighborSiteId)) == regionNeighbors

def test_to_json():
    regionJson = json.loads(repr(testVoronoiRegion))

    assert len(regionJson.keys()) == 2

    assert regionJson["siteId"] == str(testVoronoiRegion.siteId)
    assert regionJson["edges"] == [json.loads(repr(testVoronoiEdge)), json.loads(repr(testOtherVoronoiEdge))]


from uuid import uuid4

from ..edges.VoronoiEdge import VoronoiEdge

import json

def test_to_json():
    identifier0 = uuid4()
    identifier1 = uuid4()
    neighborIdentifier = uuid4()

    voronoiEdge = VoronoiEdge(vertex0Id = identifier0, vertex1Id = identifier1, neighborSiteId = neighborIdentifier)
    voronoiEdgeJson = json.loads(repr(voronoiEdge))

    assert len(voronoiEdgeJson.keys()) == 3

    assert voronoiEdgeJson["vertex0Id"] == str(identifier0)
    assert voronoiEdgeJson["vertex1Id"] == str(identifier1)
    assert voronoiEdgeJson["neighborSiteId"] == str(neighborIdentifier)


from json import dump as writeJsonOut

from .VoronoiDiagram import VoronoiDiagram
from .jsonOut.VoronoiJSONEncoder import VoronoiJSONEncoder

def toJson(voronoiDiagram: VoronoiDiagram, voronoiJsonPath: str):
    with open(voronoiJsonPath, "w") as jsonOut:
        writeJsonOut(obj = voronoiDiagram, fp = jsonOut, cls = VoronoiJSONEncoder)


from dataclasses import dataclass
from uuid import uuid4

from ..edges.VoronoiEdge import VoronoiEdge

@dataclass(frozen=True)
class VoronoiRegion:
    siteId: uuid4
    edges: tuple[VoronoiEdge]

    def neighbors(self) -> tuple[uuid4]:
        return tuple((edge.neighborSiteId for edge in self.edges))
    
    def __repr__(self) -> str:
        edges = ",".join(tuple(repr(edge) for edge in self.edges))
        return f'{{"siteId": "{str(self.siteId)}", "edges": [{edges}]}}'


from ...Point import Point
from ...VoronoiDiagram import VoronoiDiagram

from ..VoronoiJSONEncoder import VoronoiJSONEncoder

# Derived from the default points at https://cfbrasz.github.io/Voronoi.html.
testPoints = tuple((Point(x = 0.5, y = 0.8667), Point(x = 0.1667, y = 0.7222), Point(x = 0.4444, y = 0.9000)))

def test_encoding_voronoi_diagram():
    voronoiDiagram = VoronoiDiagram(basePoints = testPoints, planeWidth = 600, planeHeight = 600)
    voronoiJsonEncoder = VoronoiJSONEncoder()

    voronoiJson = voronoiJsonEncoder.default(voronoiDiagram)

    # voronout/tests suites cover the correctness of data - this just verifies the JSON object's structure.
    voronoiJsonKeys = voronoiJson.keys()

    assert len(voronoiJsonKeys) == 3

    assert 'points' in voronoiJsonKeys
    assert 'vertices' in voronoiJsonKeys
    assert 'regions' in voronoiJsonKeys


from ..Point import Point

from ..VoronoiDiagram import VoronoiDiagram
from ..VoronoiDiagramToJSON import toJson

testOutputFile = "voronoi.json"

def test_voronoi_diagram_to_json(tmp_path):
    diagramPoints = tuple(
        (Point(x = .0556, y = .1333),
        Point(x = .1667, y = .2778),
        Point(x = .4444, y = .1000))
    )

    # Other tests check VoronoiDiagram correctness - we just need to verify that the file gets created.
    voronoiJsonPath = (tmp_path / testOutputFile)
    assert not voronoiJsonPath.exists()
    
    voronoiDiagram = VoronoiDiagram(basePoints = diagramPoints, planeWidth = 600, planeHeight = 600)
    toJson(voronoiDiagram = voronoiDiagram, voronoiJsonPath = voronoiJsonPath)

    assert voronoiJsonPath.exists()


from dataclasses import dataclass
from uuid import uuid4

@dataclass(frozen=True)
class VoronoiEdge:
    vertex0Id: uuid4
    vertex1Id: uuid4
    # The region neighboring the one this is contained in.
    neighborSiteId: uuid4

    def __repr__(self) -> str:
        return f'{{"vertex0Id": "{str(self.vertex0Id)}", "vertex1Id": "{str(self.vertex1Id)}", "neighborSiteId": "{str(self.neighborSiteId)}"}}'


from ..Point import Point
from ..VoronoiDiagram import VoronoiDiagram

from pytest import raises
from uuid import uuid4

planeWidth = 600
planeHeight = 600

# All sites (and expected values) calculated via https://cfbrasz.github.io/Voronoi.html (x = "50 150 400", y = "120 250 90")
siteOne = Point(x = 0.0556, y = 0.1333)
siteTwo = Point(x = 0.1667, y = 0.2778)
siteThree = Point(x = 0.4444, y = 0.1000)

scaledSiteOne = siteOne.scale(widthScalar = planeWidth, heightScalar = planeHeight)
scaledSiteTwo = siteTwo.scale(widthScalar = planeWidth, heightScalar = planeHeight)
scaledSiteThree = siteThree.scale(widthScalar = planeWidth, heightScalar = planeHeight)

planeWidth = 600
planeHeight = 600

# For each region in testPoints, the neighbors are the other points' regions.
sitesExpectedNeighbors = {
    scaledSiteOne: tuple((scaledSiteTwo, scaledSiteThree)),
    scaledSiteTwo: tuple((scaledSiteOne, scaledSiteThree)),
    scaledSiteThree: tuple((scaledSiteOne, scaledSiteTwo))
}

expectedDiagramVertex = Point(x = 149.16, y = 59.94)

expectedBoundaryVertex1 = Point(x = 0, y = 174.6)
expectedBoundaryVertex2 = Point(x = 144.12, y = 0)
expectedBoundaryVertex3 = Point(x = 494.46, y = 600)

def _idsByPoint(pointsById: dict[uuid4, Point]) -> dict[Point, uuid4]:
    return {point: pointId for (pointId, point) in pointsById.items()}

# Verifies that VoronoiDiagram is constructed as expected.
def test_voronoi_diagram():
    testPoints = tuple((siteOne, siteTwo, siteThree))
    voronoiDiagram = VoronoiDiagram(basePoints = testPoints, planeWidth = planeWidth, planeHeight = planeHeight)

    # We should have one diagram vertex and three boundary vertices.
    assert len(voronoiDiagram.vertices) == 4

    vertexValues = voronoiDiagram.vertices.values()

    assert expectedDiagramVertex in vertexValues
    assert expectedBoundaryVertex1 in vertexValues
    assert expectedBoundaryVertex2 in vertexValues
    assert expectedBoundaryVertex3 in vertexValues
    
    # Identifiers are randomly generated, so we need to dynamically get them.
    voronoiDiagramPointsIdMap = _idsByPoint(pointsById = voronoiDiagram.points)
    
    # Check that we only have the points we passed in..
    assert len(voronoiDiagramPointsIdMap) == len(testPoints)

    scaledTestPoints = tuple((scaledSiteOne, scaledSiteTwo, scaledSiteThree))

    # .. and then check the data in terms of each point.
    for testPoint in scaledTestPoints:
        testPointIdentifier = voronoiDiagramPointsIdMap[testPoint]
        testPointRegion = voronoiDiagram.voronoiRegions[testPointIdentifier]

        assert testPointRegion.siteId == testPointIdentifier
        
        testPointRegionNeighbors = set(testPointRegion.neighbors())

        testExpectedNeighbors = sitesExpectedNeighbors[testPoint]
        testExpectedIdentifiers = set((voronoiDiagramPointsIdMap[testExpectedNeighbor] for testExpectedNeighbor in testExpectedNeighbors))

        assert testPointRegionNeighbors == testExpectedIdentifiers

def test_voronoi_diagram_too_few_base_points():
    with raises(ValueError):
        VoronoiDiagram(basePoints = (siteOne, siteTwo), planeWidth = planeWidth, planeHeight = planeHeight)

def test_voronoi_diagram_base_points_outside_bounds():
    outOfBoundsSiteThree = Point(x = 0.4444, y = 1.1000)
    with raises(ValueError):
        VoronoiDiagram(basePoints = (siteOne, siteTwo, outOfBoundsSiteThree), planeWidth = planeWidth, planeHeight = planeHeight)


from ..Point import Point
from ..Boundary import Boundary

import pytest


def test_bound_vertex_on_x_left():
    testFirstPoint = Point(x = 0.3, y = 0.4)
    testSecondPoint = Point(x = -0.3, y = 0.4)
    boundedPoint = Boundary.boundVertexOnX(vertex = testFirstPoint, otherVertex = testSecondPoint)

    assert boundedPoint.x == 0
    assert boundedPoint.y == 0.4

def test_bound_vertex_on_x_right():
    testFirstPoint = Point(x = 0.7, y = 0.4)
    testSecondPoint = Point(x = 1.3, y = 0.4)
    boundedPoint = Boundary.boundVertexOnX(vertex = testFirstPoint, otherVertex = testSecondPoint)

    assert boundedPoint.x == 1
    assert boundedPoint.y == 0.4

def test_bound_vertex_on_y_top():
    testFirstPoint = Point(x = 0.3, y = 0.7)
    testSecondPoint = Point(x = 0.3, y = 1.3)
    boundedPoint = Boundary.boundVertexOnY(vertex = testFirstPoint, otherVertex = testSecondPoint)

    assert boundedPoint.x == 0.3
    assert boundedPoint.y == 1

def test_bound_vertex_on_y_bottom():
    testFirstPoint = Point(x = 0.3, y = 0.3)
    testSecondPoint = Point(x = 0.6, y = -0.3)
    boundedPoint = Boundary.boundVertexOnY(vertex = testFirstPoint, otherVertex = testSecondPoint)

    assert boundedPoint.x == 0.45
    assert boundedPoint.y == 0

def test_bound_vertex_on_y_bottom_dx_zero():
    testFirstPoint = Point(x = 0.3, y = 0.3)
    testSecondPoint = Point(x = 0.3, y = 0.6)
    boundedPoint = Boundary.boundVertexOnY(vertex = testFirstPoint, otherVertex = testSecondPoint)

    assert boundedPoint.x == 0.3
    assert boundedPoint.y == 0

findBoundaryTestData = [
    [Point(x = 0.6, y = 0.4), Boundary.RIGHT], # Quadrant 1, angle < 45 degrees (counter-clockwise from east)
    [Point(x = 0.4, y = 0.6), Boundary.TOP], # Quadrant 1, 45 degrees < angle < 90 degrees
    [Point(x = -0.4, y = 0.6), Boundary.TOP], # Quadrant 2, 90 degrees < angle < 135 degrees
    [Point(x = -0.6, y = 0.4), Boundary.LEFT], # Quadrant 2, 135 degrees < angle < 180 degrees
    [Point(x = -0.6, y = -0.4), Boundary.LEFT], # Quadrant 3, 180 degrees < angle < 225 degrees
    [Point(x = -0.4, y = -0.6), Boundary.BOTTOM], # Quadrant 3, 225 degrees < angle < 270 degres
    [Point(x = 0.4, y = -0.6), Boundary.BOTTOM], # Quadrant 4, 270 degrees < angle < 315 degrees
    [Point(x = 0.6, y = -0.4), Boundary.RIGHT] # Quadrant 4, 315 degrees < angle < 360 degrees
]

@pytest.mark.parametrize("linePoint, expectedBoundary", findBoundaryTestData)
def test_find_boundary_in_line_direction(linePoint: Point, expectedBoundary: Boundary):
    quadrantPoint = Point(x = 0, y = 0)
    assert Boundary.findBoundaryInLineDirection(linePoint1 = quadrantPoint, linePoint2 = linePoint) == expectedBoundary

def test_find_boundary_line_intersection_point():
    intersectionPoint = Boundary.boundaryLineIntersectionPoint(lineFirstPoint = Point(x = 0.3, y = 0.4), lineSecondPoint = Point(x = -0.1, y = 0), boundary = Boundary.LEFT)
    
    # (0.3, 0.4) -> (-0.1, 0) would intersect at (0, 0.1)
    assert intersectionPoint.x == 0
    assert intersectionPoint.y == 0.1


from enum import Enum
from .Point import Point

from math import acos, degrees

from numpy import array as npArray, dot as npDotProduct

from .utils import boundValue

# _Quadrant is, on a line (p1, p2), with respect to an x/y graph with p1 at 0, 0
class _Quadrant(Enum):
    FIRST = 1 # dx > 0, dy > 0
    SECOND = 2 # dx < 0, dy > 0
    THIRD = 3 # dx < 0, dy < 0
    FOURTH = 4 # dx > 0, dy < 0

class Boundary(Enum):
    TOP = 1
    RIGHT = 2
    BOTTOM = 3
    LEFT = 4

    @staticmethod
    def _calculateLineSlope(linePoint1: Point, linePoint2: Point) -> float:
        return (linePoint2.y - linePoint1.y)/(linePoint2.x - linePoint1.x)

    # The boundary (line) a VoronoiDiagram edge might intersect with.
    @staticmethod
    def _getBoundaryLine(boundary: Boundary) -> tuple[Point]:
        match boundary:
            case Boundary.TOP: return tuple((Point(x = 0, y = 1), Point(x = 1, y = 1)))
            case Boundary.RIGHT: return tuple((Point(x = 1, y = 0), Point(x = 1, y = 1)))
            case Boundary.BOTTOM: return tuple((Point(x = 0, y = 0), Point(x = 1, y = 0)))
            case Boundary.LEFT: return tuple((Point(x = 0, y = 0), Point(x = 0, y = 1)))
    
    # quadrantVector approximates the relevant x/y-axis segment.
    @staticmethod
    def _getQuadrantVector(quadrant: _Quadrant, point: Point) -> npArray[float]:
        match quadrant:
            case _Quadrant.FIRST: return npArray((1, point.y))
            case _Quadrant.SECOND: return npArray((point.x, 1))
            case _Quadrant.THIRD: return npArray((-1, point.y))
            case _Quadrant.FOURTH: return npArray((point.x, -1))

    # Returns the next (counter-clockwise from quadrant) _Quadrant.
    @staticmethod
    def _determineNextQuadrant(quadrant: _Quadrant) -> _Quadrant:
        match quadrant:
            case _Quadrant.FIRST: return _Quadrant.SECOND
            case _Quadrant.SECOND: return _Quadrant.THIRD
            case _Quadrant.THIRD: return _Quadrant.FOURTH
            case _Quadrant.FOURTH: return _Quadrant.FIRST

    # https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line#Line_defined_by_two_points
    @staticmethod
    def _pointDistanceToBoundary(point: Point, boundary: Boundary) -> float:
        (boundaryFirstPoint, boundarySecondPoint) = Boundary._getBoundaryLine(boundary)

        secondFirstDx = boundarySecondPoint.x - boundaryFirstPoint.x
        secondFirstDy = boundarySecondPoint.y - boundaryFirstPoint.y

        x2y1 = boundarySecondPoint.x * boundaryFirstPoint.y
        y2x1 = boundarySecondPoint.y * boundaryFirstPoint.x

        distanceNumerator = abs((secondFirstDy * point.x) - (secondFirstDx * point.y) + x2y1 - y2x1)
        distanceDenominator = Point.distance(p1 = boundaryFirstPoint, p2 = boundarySecondPoint)
                
        return distanceNumerator / distanceDenominator

    @staticmethod
    def _findClosestBoundaryToPoint(point: Point, boundaries: Boundary) -> Boundary:
        return sorted(boundaries, key = lambda boundary: Boundary._pointDistanceToBoundary(point = point, boundary = boundary))[0]

    # Like Point.distance(Point(vx, vy), Point(0, 0)), but I don't want to pass Point(0, 0).    
    @staticmethod
    def _calculateVectorMagnitude(vectorPoint: Point) -> float:
        return Point.distance(p1 = vectorPoint, p2 = Point(x = 0, y = 0))
    
    # Quadrants on an x/y graph are counterclockwise from east.
    @staticmethod
    def _determineVectorQuadrant(originatingPoint: Point, vectorPoint: Point) -> _Quadrant:
        dx = vectorPoint.x - originatingPoint.x
        dy = vectorPoint.y - originatingPoint.y

        if dx > 0 and dy > 0:
            return _Quadrant.FIRST
        elif dx < 0 and dy > 0:
            return _Quadrant.SECOND
        elif dx < 0 and dy < 0:
            return _Quadrant.THIRD
        else: # dx > 0 and dy < 0
            return _Quadrant.FOURTH
    
    # https://www.cuemath.com/geometry/angle-between-vectors/
    @staticmethod
    def _calculatePointQuadrantVectorsAngle(quadrantVectorPoint: Point, vectorPoint: Point, quadrant: _Quadrant) -> float:
        pointVector = npArray((vectorPoint.x, vectorPoint.y))
        quadrantVector = Boundary._getQuadrantVector(quadrant = quadrant, point = quadrantVectorPoint)

        vectorDotProduct = npDotProduct(pointVector, quadrantVector)
        
        pointMagnitude = Boundary._calculateVectorMagnitude(vectorPoint = vectorPoint)
        quadrantMagnitude = Boundary._calculateVectorMagnitude(vectorPoint = Point(x = quadrantVector[0], y = quadrantVector[1]))

        calculatedCos = vectorDotProduct / (pointMagnitude * quadrantMagnitude)

        # acos expects values in the range of [-1, 1]
        if calculatedCos > 0:
            return degrees(acos(min(calculatedCos, 1)))
        else:
            # calculatedCos <= 0
            return degrees(acos(max(calculatedCos, -1)))
    
    @staticmethod
    def findBoundaryInLineDirection(linePoint1: Point, linePoint2: Point) -> Boundary:
        lineQuadrant = Boundary._determineVectorQuadrant(originatingPoint = linePoint1, vectorPoint = linePoint2)
        nextQuadrant = Boundary._determineNextQuadrant(quadrant = lineQuadrant)

        lineQuadrantAngle = Boundary._calculatePointQuadrantVectorsAngle(quadrantVectorPoint = linePoint1, vectorPoint = linePoint2, quadrant = lineQuadrant)
        nextQuadrantAngle = Boundary._calculatePointQuadrantVectorsAngle(quadrantVectorPoint = linePoint1, vectorPoint = linePoint2, quadrant = nextQuadrant)

        match lineQuadrant:
            case _Quadrant.FIRST: return Boundary.RIGHT if lineQuadrantAngle < nextQuadrantAngle else Boundary.TOP
            case _Quadrant.SECOND: return Boundary.TOP if lineQuadrantAngle < nextQuadrantAngle else Boundary.LEFT
            case _Quadrant.THIRD: return Boundary.LEFT if lineQuadrantAngle < nextQuadrantAngle else Boundary.BOTTOM
            case _Quadrant.FOURTH: return Boundary.BOTTOM if lineQuadrantAngle < nextQuadrantAngle else Boundary.RIGHT
 
    @staticmethod
    # https://en.wikipedia.org/wiki/Line%E2%80%93line_intersection#Given_two_points_on_each_line
    def boundaryLineIntersectionPoint(lineFirstPoint: Point, lineSecondPoint: Point, boundary: Boundary) -> Point:
        (boundaryFirstPoint, boundarySecondPoint) = Boundary._getBoundaryLine(boundary)

        x2dx1 = lineSecondPoint.x - lineFirstPoint.x
        y2dy1 = lineSecondPoint.y - lineFirstPoint.y

        # If line has a definite slope, we calculate intersection - if not, we just figure out which boundary it would vertically/horizontally intersect.
        if x2dx1 and y2dy1:
            x1y2 = lineFirstPoint.x * lineSecondPoint.y
            y1x2 = lineFirstPoint.y * lineSecondPoint.x

            x3dx4 = boundaryFirstPoint.x - boundarySecondPoint.x
            x1dx2 = -x2dx1

            x3y4 = boundaryFirstPoint.x * boundarySecondPoint.y
            y3x4 = boundaryFirstPoint.y * boundarySecondPoint.x

            y1dy2 = -y2dy1
            y3dy4 = boundaryFirstPoint.y - boundarySecondPoint.y

            pointDenominator = (x1dx2 * y3dy4) - (y1dy2 * x3dx4)

            pointXNumerator = ((x1y2 - y1x2) * x3dx4) - (x1dx2 * (x3y4 - y3x4))

            intersectionX = pointXNumerator / pointDenominator

            pointYNumerator = ((x1y2 - y1x2) * y3dy4) - (y1dy2 * (x3y4 - y3x4))

            intersectionY = pointYNumerator / pointDenominator

            intersectionPoint = Point(x = boundValue(value = intersectionX), y = boundValue(value = intersectionY))

            return intersectionPoint
        else:
            if x2dx1 and not y2dy1:
                # Lines with only dx will intersect either Left or Right boundaries.
                xCoord = 0 if lineSecondPoint.x < lineFirstPoint.x else 1
                xIntersectionPoint = Point(x = xCoord, y = lineSecondPoint.y)
                return xIntersectionPoint
            elif y2dy1 and not x2dx1:
                # Lines with only dy will intersect either Top or Bottom boundaries.
                yCoord = 0 if lineSecondPoint.y < lineFirstPoint.y else 1
                yIntersectionPoint = Point(x = lineSecondPoint.x, y = yCoord)
                return yIntersectionPoint
            else:
                # Lines without dx or dy shouldn't happen. 
                raise ValueError(f"Line {lineFirstPoint}, {lineSecondPoint} unexpectedly has both dx and dy = 0")
            
    @staticmethod
    def boundVertexOnX(vertex: Point, otherVertex: Point) -> Point:
        closestXBoundary = Boundary._findClosestBoundaryToPoint(point = vertex, boundaries = tuple((Boundary.LEFT, Boundary.RIGHT)))

        # xBound can only be Left or Right
        xBound = 0 if closestXBoundary == Boundary.LEFT else 1

        # Calculate (dy/dx) and dx..
        verticesSlope = Boundary._calculateLineSlope(linePoint1 = otherVertex, linePoint2 = vertex)
        boundDx = xBound - otherVertex.x

        # .. to do updatedY = otherVertex.y + dy.
        verticesSlopeDx = verticesSlope * boundDx
        updatedY = verticesSlopeDx + otherVertex.y

        boundedOnX = Point(x = boundValue(value = xBound), y = boundValue(value = updatedY))
        return boundedOnX
    
    @staticmethod
    def boundVertexOnY(vertex: Point, otherVertex: Point) -> Point:
        closestYBoundary = Boundary._findClosestBoundaryToPoint(point = vertex, boundaries = tuple((Boundary.BOTTOM, Boundary.TOP)))
        # yBound can only be Bottom or Top
        yBound = 0 if closestYBoundary == Boundary.BOTTOM else 1

        if vertex.x != otherVertex.x:
            # Calculate (dy/dx) and dy..
            verticesSlope = Boundary._calculateLineSlope(linePoint1 = otherVertex, linePoint2 = vertex)
            boundDy = yBound - otherVertex.y

            # .. to do updatedX = otherVertex.x + dx.
            verticesSlopeDy = boundDy/verticesSlope
            updatedX = verticesSlopeDy + otherVertex.x

            boundedOnY = Point(x = boundValue(value = updatedX), y = boundValue(value = yBound))
            return boundedOnY
        else:
            # We shouldn't calculate any slope if dx = 0.
            return Point(x = vertex.x, y = yBound)


from decimal import Decimal, ROUND_HALF_EVEN

_bound = Decimal('0.0001')

def boundValue(value: float) -> float:
    # bound to 4 places with as tight rounding as possible
    return float(Decimal(value).quantize(exp = _bound, rounding = ROUND_HALF_EVEN))


from .. import boundValue

def test_bounding_over_bound():
    # bound = 0.0001, rounding = ROUND_HALF_EVEN
    assert boundValue(value = 1.23456) == 1.2346

def test_bounding_under_bound():
    assert boundValue(value = 1.23) == 1.23


from dataclasses import dataclass, field
from uuid import uuid4

from ..edges.VoronoiEdgeData import VoronoiEdgeData

@dataclass
class VoronoiRegionData:
    siteId: uuid4
    # A dict of edges such that iterating over it will make one complete loop around the region. .. str needs to be a key that can handle multiple [0, -1] cases
    _edgesData: dict[str, VoronoiEdgeData] = field(default_factory=dict)

    def _makeEdgeVerticesIdentifier(self, edgeVertex0Index: int, edgeVertex1Index: int, neighborSiteIndex: int) -> str:
        return f"{edgeVertex0Index}_{edgeVertex1Index}_{neighborSiteIndex}"
    
    def addEdgesData(self, edgeVertex0Index: int, edgeVertex1Index: int, neighborSiteIndex: int, edgeData: VoronoiEdgeData) -> bool:
        edgeVerticies01Identifier = self._makeEdgeVerticesIdentifier(edgeVertex0Index = edgeVertex0Index, edgeVertex1Index = edgeVertex1Index, neighborSiteIndex = neighborSiteIndex)
        edgeVerticies10Identifier = self._makeEdgeVerticesIdentifier(edgeVertex0Index = edgeVertex1Index, edgeVertex1Index = edgeVertex0Index, neighborSiteIndex = neighborSiteIndex)

        edgeDataAlreadyAdded = edgeVerticies01Identifier in self._edgesData or edgeVerticies10Identifier in self._edgesData
        if not edgeDataAlreadyAdded:
            # .. then it's OK to add self._edgesData[edgeVerticies01Identifier].
            self._edgesData[edgeVerticies01Identifier] = edgeData
            return True
        else:
            # .. then it's not OK to add self._edgesData[edgeVerticies01Identifier], and we should indicate that.
            return False

    def edges(self):
        for edgeData in self._edgesData.values():
            yield edgeData


from dataclasses import dataclass
from uuid import uuid4

from ..Point import Point
from .VoronoiEdge import VoronoiEdge

@dataclass(frozen=True)
class VoronoiEdgeData:
    vertex0: Point
    vertex1: Point

    edgeInIds: VoronoiEdge

    def vertex0Id(self) -> uuid4:
        return self.edgeInIds.vertex0Id
    
    def vertex1Id(self) -> uuid4:
        return self.edgeInIds.vertex1Id
    
    def neighborSiteId(self) -> uuid4:
        return self.edgeInIds.neighborSiteId


from .Boundary import Boundary
from .Point import Point

from .VoronoiDiagram import VoronoiDiagram

from .edges.VoronoiEdge import VoronoiEdge
from .edges.VoronoiEdgeData import VoronoiEdgeData

from .regions.VoronoiRegion import VoronoiRegion

from .jsonOut import VoronoiJSONEncoder
from .VoronoiDiagramToJSON import toJson


from json import JSONEncoder, loads as loadJSONString
from uuid import UUID

from ..Point import Point
from ..VoronoiDiagram import VoronoiDiagram

class VoronoiJSONEncoder(JSONEncoder):
    def _handlePointDict(self, pointDict: dict[UUID, Point]) -> dict[str, dict[str, float]]:
        return {str(key): loadJSONString(repr(value)) for (key, value) in pointDict.items()}
    
    def default(self, obj):
        if isinstance(obj, VoronoiDiagram):
            return {
                'points': self._handlePointDict(obj.points),
                'vertices': self._handlePointDict(obj.vertices),
                'regions': tuple((loadJSONString(repr(region)) for (_, region) in obj.voronoiRegions.items()))
            }
        else:
            return super().default(obj)


from .Boundary import Boundary
from .Point import Point

from .edges.VoronoiEdge import VoronoiEdge

from .edges.VoronoiEdgeData import VoronoiEdgeData

from .regions.VoronoiRegion import VoronoiRegion
from .regions.VoronoiRegionData import VoronoiRegionData

from enum import Enum

from itertools import repeat

from scipy.spatial import Voronoi
from uuid import uuid4

import numpy as np

class _BoundingNeeded(Enum):
    X_AND_Y = 1
    X = 2
    Y = 3

# Minimum basePoints that the underlying qHull requires.
minBasePoints = 3

class VoronoiDiagram:
    def __init__(self, basePoints: tuple[Point], planeWidth: float, planeHeight: float):
        # Make sure basePoints fits the minBasePoints/within (0, 0) -> (1, 1) constraints.
        self._validateBasePoints(basePoints = basePoints)

        # We expect basePoints to have 0, 0 (top-left), but scipy.spatial does 0, 0 (bottom-left) - so convert.
        sciPySpatialPoints = np.array(tuple(basePoint.convertPointBase() for basePoint in basePoints))        
        self._voronoiDiagram = Voronoi(sciPySpatialPoints)

        # _spatial.. values are with respect to the diagram. 
        self._spatialSites = { uuid4(): Point(x = spatialPoint[0], y = spatialPoint[1]) for spatialPoint in self._voronoiDiagram.points }
        self._spatialSiteKeys = tuple(self._spatialSites.keys())

        self._spatialDiagramVertices = { uuid4(): Point(x = spatialDiagramVertex[0], y = spatialDiagramVertex[1]) for spatialDiagramVertex in self._voronoiDiagram.vertices}
        spatialDiagramVerticesKeys = list(self._spatialDiagramVertices.keys())

        # { regionId: <list of VoronoiEdgeIdData describing the edges that bound the region> }
        self._spatialSiteRegionBoundaries: dict[uuid4, list[VoronoiEdge]] = { siteKey: [] for siteKey in self._spatialSiteKeys }

        # Stores vertices determined as a result of " bounding " calculations.
        self._spatialBoundingVertices: dict[uuid4, Point] = {}

        # Store vertices that get bounded for later deletion.
        boundedDiagramVertices = []

        # voronoiRegionsInfo[pointIndex] describes the region identified by self._spatialSitesKeys[pointIndex].
        voronoiRegionsInfo = tuple((self._makeVoronoiRegionData(regionSiteIdIndex = pointIndex) for pointIndex in range(len(self._voronoiDiagram.points))))
        
        # Handling len(basePoints) = 3 case where Voronoi diagram is created with only one vertex.
        numBasePoints = len(basePoints)
        vertices = self._voronoiDiagram.ridge_vertices
        verticesToZip = vertices if numBasePoints > minBasePoints else tuple(repeat(vertices[0], numBasePoints))
        
        # Each ([vertex0, vertex1], [site0, site1]) should concern the same edge.
        edgeVerticesAndSites = () if False else zip(verticesToZip, self._voronoiDiagram.ridge_points)

        for ([edgeVertex0Index, edgeVertex1Index], [edgeSite0Index, edgeSite1Index]) in edgeVerticesAndSites:
            vertex0OutOfDiagram = edgeVertex0Index == -1
            vertex1OutOfDiagram = edgeVertex1Index == -1

            # None indicates that the relevant vertex is outside the diagram's bounds and must be calculated.
            edgeVertex0Id = None if vertex0OutOfDiagram else spatialDiagramVerticesKeys[edgeVertex0Index]
            edgeVertex1Id = None if vertex1OutOfDiagram else spatialDiagramVerticesKeys[edgeVertex1Index]

            edgeVertex0 = None if vertex0OutOfDiagram else self._spatialDiagramVertices[edgeVertex0Id]
            edgeVertex1 = None if vertex1OutOfDiagram else self._spatialDiagramVertices[edgeVertex1Id]

            edgeSite0Id = self._spatialSiteKeys[edgeSite0Index]
            edgeSite1Id = self._spatialSiteKeys[edgeSite1Index]

            # Add edge info to both regions containing the edge.
            site0Edge = VoronoiEdge(vertex0Id = edgeVertex0Id, vertex1Id = edgeVertex1Id, neighborSiteId = edgeSite1Id)
            site0EdgeData = VoronoiEdgeData(vertex0 = edgeVertex0, vertex1 = edgeVertex1, edgeInIds = site0Edge)
            voronoiRegionsInfo[edgeSite0Index].addEdgesData(edgeVertex0Index = edgeVertex0Index, edgeVertex1Index = edgeVertex1Index, neighborSiteIndex = edgeSite1Index, edgeData = site0EdgeData)
            
            site1Edge = VoronoiEdge(vertex0Id = edgeVertex0Id, vertex1Id = edgeVertex1Id, neighborSiteId = edgeSite0Id)
            site1EdgeData = VoronoiEdgeData(vertex0 = edgeVertex0, vertex1 = edgeVertex1, edgeInIds = site1Edge)
            voronoiRegionsInfo[edgeSite1Index].addEdgesData(edgeVertex0Index = edgeVertex1Index, edgeVertex1Index = edgeVertex0Index, neighborSiteIndex = edgeSite0Index, edgeData = site1EdgeData)

        for voronoiRegionInfo in voronoiRegionsInfo:
            regionId = voronoiRegionInfo.siteId
            for voronoiRegionEdge in voronoiRegionInfo.edges():
                edgeVertex0 = voronoiRegionEdge.vertex0
                edgeVertex1 = voronoiRegionEdge.vertex1

                edgeVertex0Id = voronoiRegionEdge.vertex0Id()
                edgeVertex1Id = voronoiRegionEdge.vertex1Id()

                vertex0OutsideDiagram = not voronoiRegionEdge.vertex0 and not voronoiRegionEdge.vertex0Id()
                vertex1OutsideDiagram = not voronoiRegionEdge.vertex1 and not voronoiRegionEdge.vertex1Id()

                neighborSiteId = voronoiRegionEdge.neighborSiteId()

                if not (vertex0OutsideDiagram or vertex1OutsideDiagram):
                    vertex0NeedsBounding = self._vertexNeedsBounding(vertex = edgeVertex0)
                    vertex1NeedsBounding = self._vertexNeedsBounding(vertex = edgeVertex1)

                    maybeBoundedEdgeVertex0 = self._boundVertex(vertex = edgeVertex0, otherVertex = edgeVertex1, boundingNeeded = vertex0NeedsBounding) if vertex0NeedsBounding else None
                    latestEdgeVertex0 = maybeBoundedEdgeVertex0 or edgeVertex0

                    edgeVertex0Id = voronoiRegionEdge.vertex0Id()
                    if maybeBoundedEdgeVertex0 and edgeVertex0Id not in boundedDiagramVertices:
                        boundedDiagramVertices.append(edgeVertex0Id)

                    edgeVertex1Id = voronoiRegionEdge.vertex1Id()
                    maybeBoundedEdgeVertex1 = self._boundVertex(vertex = edgeVertex1, otherVertex = latestEdgeVertex0, boundingNeeded = vertex1NeedsBounding) if vertex1NeedsBounding else None
                    latestEdgeVertex1 = maybeBoundedEdgeVertex1 or edgeVertex1

                    if maybeBoundedEdgeVertex1 and edgeVertex1Id not in boundedDiagramVertices:
                        boundedDiagramVertices.append(edgeVertex1Id)

                    latestEdgeVertex0Id = self._getBoundingVertexId(boundingVertex = maybeBoundedEdgeVertex0) if maybeBoundedEdgeVertex0 else edgeVertex0Id

                    if maybeBoundedEdgeVertex0:
                        self._spatialBoundingVertices[latestEdgeVertex0Id] = latestEdgeVertex0
                    
                    latestEdgeVertex1Id = self._getBoundingVertexId(boundingVertex = maybeBoundedEdgeVertex1) if maybeBoundedEdgeVertex1 else edgeVertex1Id

                    if maybeBoundedEdgeVertex1:
                        self._spatialBoundingVertices[latestEdgeVertex1Id] = latestEdgeVertex1

                    latestEdge = VoronoiEdge(vertex0Id = latestEdgeVertex0Id, vertex1Id = latestEdgeVertex1Id, neighborSiteId = neighborSiteId)
                    self._spatialSiteRegionBoundaries[regionId].append(latestEdge)
                elif vertex0OutsideDiagram or vertex1OutsideDiagram:
                    pointInsideDiagramId = edgeVertex0Id if vertex1OutsideDiagram else edgeVertex1Id
                    
                    pointInsideDiagram = self._spatialDiagramVertices[pointInsideDiagramId]
                    pointInsideDiagramNeedsBounding = self._vertexNeedsBounding(vertex = pointInsideDiagram)

                    regionSite = self._spatialSites[regionId]
                    neighborSite = self._spatialSites[neighborSiteId]

                    betweenRegionsMidpoint = Point.midpoint(p1 = regionSite, p2 = neighborSite)
                    maybeRegionContainingCalculatedMidpoint = self._maybeReturnSiteOfRegionContainingMidpoint(calculatedMidpoint = betweenRegionsMidpoint, calculationPoint1 = regionSite, calculationPoint2 = neighborSite)
                    # Sometimes, the calculated midpoint is within a region. We try to avoid this by reflecting it over the closest vertex.
                    reflectedSitesMidpoint = None if not maybeRegionContainingCalculatedMidpoint else self._reflectPointAroundVertex(point = betweenRegionsMidpoint, vertex = pointInsideDiagram)

                    latestMidpoint = reflectedSitesMidpoint or betweenRegionsMidpoint

                    maybeBoundedPointInsideDiagram = self._boundVertex(vertex = pointInsideDiagram, otherVertex = latestMidpoint, boundingNeeded = pointInsideDiagramNeedsBounding) if pointInsideDiagramNeedsBounding else None
                    latestPointInsideDiagram = maybeBoundedPointInsideDiagram or pointInsideDiagram

                    if maybeBoundedPointInsideDiagram and pointInsideDiagramId not in boundedDiagramVertices:
                        boundedDiagramVertices.append(pointInsideDiagramId)

                    closestBoundary = Boundary.findBoundaryInLineDirection(linePoint1 = latestPointInsideDiagram, linePoint2 = latestMidpoint)
                    boundaryIntersectionPoint = Boundary.boundaryLineIntersectionPoint(lineFirstPoint = latestPointInsideDiagram, lineSecondPoint = latestMidpoint, boundary = closestBoundary)

                    boundaryIntersectionNeedsBounding = self._vertexNeedsBounding(vertex = boundaryIntersectionPoint)
                    maybeBoundedIntersection = self._boundVertex(vertex = boundaryIntersectionPoint, otherVertex = latestPointInsideDiagram, boundingNeeded = boundaryIntersectionNeedsBounding) if boundaryIntersectionNeedsBounding else None

                    latestPointInsideDiagramId = self._getBoundingVertexId(boundingVertex = maybeBoundedPointInsideDiagram) if maybeBoundedPointInsideDiagram else pointInsideDiagramId

                    if maybeBoundedPointInsideDiagram:
                        self._spatialBoundingVertices[latestPointInsideDiagramId] = latestPointInsideDiagram

                    latestBoundaryIntersection = maybeBoundedIntersection or boundaryIntersectionPoint        
                    boundaryIntersectionPointId = self._getBoundingVertexId(boundingVertex = latestBoundaryIntersection)
                    self._spatialBoundingVertices[boundaryIntersectionPointId] = latestBoundaryIntersection

                    latestEdge = VoronoiEdge(vertex0Id = boundaryIntersectionPointId, vertex1Id = latestPointInsideDiagramId, neighborSiteId = neighborSiteId) if vertex0OutsideDiagram else VoronoiEdge(edgeVertex0Id = latestPointInsideDiagramId, edgeVertex1Id = boundaryIntersectionPointId, neighborSiteId = neighborSiteId)
                    self._spatialSiteRegionBoundaries[regionId].append(latestEdge)
                else:
                    # Should only be needed for assumedly impossible cases like bertex0/1 both being outside the diagram.
                    raise ValueError(f"Could not handle unexpected vertex0/1OutsideDiagram ({vertex0OutsideDiagram}, {vertex1OutsideDiagram})")

        for boundedDiagramVertex in boundedDiagramVertices:
            del self._spatialDiagramVertices[boundedDiagramVertex]

        self.voronoiRegions = { spatialSiteKey: self._makeVoronoiRegion(regionSiteIdentifier = spatialSiteKey) for spatialSiteKey in self._spatialSiteKeys }

        # Public-facing values are 0, 0 (top-left).
        self.points = {pointId: point.convertPointBase().scale(widthScalar = planeWidth, heightScalar = planeHeight) for (pointId, point) in self._spatialSites.items()}
        
        convertedDiagramVertices = {diagramVertexId: diagramVertex.convertPointBase().scale(widthScalar = planeWidth, heightScalar = planeHeight) for (diagramVertexId, diagramVertex) in self._spatialDiagramVertices.items()}
        convertedBoundingVertices = {boundingVertexId: boundingVertex.convertPointBase().scale(widthScalar = planeWidth, heightScalar = planeHeight) for (boundingVertexId, boundingVertex) in self._spatialBoundingVertices.items()}

        self.vertices = convertedDiagramVertices | convertedBoundingVertices
        
    def _validateBasePoints(self, basePoints: tuple[Point]) -> None:
        if len(basePoints) < minBasePoints:
            raise ValueError(f"Too few points specified, {basePoints} - need minimum {minBasePoints}")
        
        allWithinBounds = all((0 <= basePoint.x <= 1 and 0 <= basePoint.y <= 1 for basePoint in basePoints))
        if not allWithinBounds:
            raise ValueError(f"{basePoints} violate the x/y must be >= 0, <= 1 constraint")

    def _makeVoronoiRegionData(self, regionSiteIdIndex: int) -> VoronoiRegionData:
        regionSiteId = self._spatialSiteKeys[regionSiteIdIndex]
        return VoronoiRegionData(siteId = regionSiteId)

    def _getBoundingVertexId(self, boundingVertex: Point) -> uuid4:
        extantIdSearchResult = tuple((vertexId for (vertexId, vertex) in self._spatialBoundingVertices.items() if vertex == boundingVertex))
        if extantIdSearchResult:
            # Return the first (and only) result.
            return extantIdSearchResult[0]
        else:
            # Return a new ID that will be stored.
            return uuid4()

    def _pointXWithinBounds(self, point: Point) -> bool:
        return 0 <= point.x <= 1
    
    def _pointYWithinBounds(self, point: Point) -> bool:
        return 0 <= point.y <= 1
    
    def _vertexNeedsBounding(self, vertex: Point) -> _BoundingNeeded | None:
        vertexXUnbounded = not self._pointXWithinBounds(point = vertex)
        vertexYUnbounded = not self._pointYWithinBounds(point = vertex)

        if vertexXUnbounded and vertexYUnbounded:
            return _BoundingNeeded.X_AND_Y
        elif vertexXUnbounded and not vertexYUnbounded:
            return _BoundingNeeded.X
        elif not vertexXUnbounded and vertexYUnbounded:
            return _BoundingNeeded.Y
        else:
            return None
        
    def _boundVertex(self, vertex: Point, otherVertex: Point, boundingNeeded: _BoundingNeeded) -> Point:
        match boundingNeeded:
            case _BoundingNeeded.X_AND_Y:
                vertexBoundedOnX = Boundary.boundVertexOnX(vertex = vertex, otherVertex = otherVertex)
                
                shouldBoundOnY = 0 <= vertexBoundedOnX.x <= 1 and not (0 <= vertexBoundedOnX.y <= 1)
                vertexBoundedOnY = Boundary.boundVertexOnY(vertex = vertexBoundedOnX, otherVertex = otherVertex) if shouldBoundOnY else None
                
                return vertexBoundedOnY or vertexBoundedOnX
            case _BoundingNeeded.X:
                return Boundary.boundVertexOnX(vertex = vertex, otherVertex = otherVertex)
            case _BoundingNeeded.Y:
                return Boundary.boundVertexOnY(vertex = vertex, otherVertex = otherVertex)
            
    def _maybeReturnSiteOfRegionContainingMidpoint(self, calculatedMidpoint: Point, calculationPoint1: Point, calculationPoint2: Point) -> Point:
        otherSites = tuple((otherSite for otherSite in self._spatialSites.values() if otherSite != calculationPoint1 and otherSite != calculationPoint2))

        calculationPoint1Dist = Point.distance(p1 = calculatedMidpoint, p2 = calculationPoint1)
        calculationPoint2Dist = Point.distance(p1 = calculatedMidpoint, p2 = calculationPoint2)

        otherSiteDistances = { otherSite: Point.distance(p1 = calculatedMidpoint, p2 = otherSite) for otherSite in otherSites }
        # Filter otherSiteDistances down to any closer to otherSite than calculationPoint1 and calculationPoint2.
        otherSiteDistancesFiltered = tuple((otherSite for (otherSite, otherSiteDistance) in otherSiteDistances.items() if otherSiteDistance < calculationPoint1Dist and otherSiteDistance < calculationPoint2Dist))
        
        if otherSiteDistancesFiltered:
            # .. return the first, very likely only, result.
            return otherSiteDistancesFiltered[0]
        else:
            # .. return None, because the midpoint is on a line.
            return None
        
    def _regionContainsMidpoint(self, midpoint: Point, regionSite: Point) -> bool:
        regionMidpointDistance = Point.distance(p1 = midpoint, p2 = regionSite)
        otherSiteDistances = tuple((Point.distance(p1 = midpoint, p2 = otherSite) for otherSite in self._spatialSites.values() if otherSite != regionSite))
        return all((regionMidpointDistance < otherSiteDistance for otherSiteDistance in otherSiteDistances))
            
    def _reflectPointAroundVertex(self, point: Point, vertex: Point) -> Point:
        pointDy = vertex.y - point.y
        pointDx = vertex.x - point.x

        reflectionY = vertex.y + pointDy
        reflectionX = vertex.x + pointDx
        
        reflection = Point(x = reflectionX, y = reflectionY)
        return reflection
    
    def _reflectSitesAndCalculateMidpoint(self, site1: Point, site2: Point, vertex: Point) -> Point:
        reflectedSite1 = self._reflectPointAroundVertex(point = site1, vertex = vertex)
        reflectedSite2 = self._reflectPointAroundVertex(point = site2, vertex = vertex)

        return Point.midpoint(p1 = reflectedSite1, p2 = reflectedSite2)
    
    def _makeVoronoiRegion(self, regionSiteIdentifier: uuid4) -> VoronoiRegion:
        regionEdges = self._spatialSiteRegionBoundaries[regionSiteIdentifier]
        return VoronoiRegion(siteId = regionSiteIdentifier, edges = regionEdges)